"""Interact with the Censys Clouds API."""

from ..common.types import Datetime
from ..common.utils import format_iso8601
from .api import CensysAsmAPI


class Clouds(CensysAsmAPI):
    """Clouds API class."""

    base_path = "/v1/clouds"

    def get_host_counts(
        self,
        since: Datetime,
    ) -> dict:
        """Retrieve host counts by cloud.

        Hosts found after the date provided in the `since` parameter will be included in the new asset counts.

        Args:
            since (Datetime): Date to include hosts from.

        Returns:
            dict: Host count result.
        """
        since = format_iso8601(since)
        return self._get(f"{self.base_path}/hostCounts/{since}")

    def get_domain_counts(self, since: Datetime) -> dict:
        """Retrieve domain counts by cloud.

        Args:
            since (Datetime): Date to include domains from.

        Returns:
            dict: Domain count result.
        """
        since = format_iso8601(since)
        return self._get(f"{self.base_path}/domainCounts/{since}")

    def get_object_store_counts(self, since: Datetime) -> dict:
        """Retrieve object store counts by cloud.

        Args:
            since (Datetime): Date to include object stores from.

        Returns:
            dict: Object store count result.
        """
        since = format_iso8601(since)
        return self._get(f"{self.base_path}/objectStoreCounts/{since}")

    def get_subdomain_counts(self, since: Datetime) -> dict:
        """Retrieve subdomain counts by cloud.

        Args:
            since (Datetime): Date to include subdomains from.

        Returns:
            dict: Subdomain count result.
        """
        since = format_iso8601(since)
        return self._get(f"{self.base_path}/subdomainCounts/{since}")

    def get_unknown_counts(self) -> dict:
        """Retrieve known and unknown counts for hosts by cloud.

        Returns:
            dict: Unknown count result.
        """
        return self._get(f"{self.base_path}/unknownCounts")



"""An easy-to-use and lightweight API wrapper for Censys ASM (app.censys.io)."""

from .assets import (
    Assets,
    CertificatesAssets,
    DomainsAssets,
    HostsAssets,
    ObjectStoragesAssets,
    SubdomainsAssets,
    WebEntitiesAssets,
)
from .beta import Beta
from .client import AsmClient
from .clouds import Clouds
from .inventory import InventorySearch
from .logbook import Events, Logbook
from .risks import Risks
from .saved_queries import SavedQueries
from .seeds import Seeds

__all__ = [
    "AsmClient",
    "Assets",
    "Beta",
    "CertificatesAssets",
    "Clouds",
    "DomainsAssets",
    "Events",
    "HostsAssets",
    "InventorySearch",
    "Logbook",
    "Risks",
    "SavedQueries",
    "Seeds",
    "SubdomainsAssets",
    "WebEntitiesAssets",
    "ObjectStoragesAssets",
]



"""Interact with the Censys Object Storage Assets API."""

from urllib.parse import quote

from .assets import Assets


class ObjectStoragesAssets(Assets):
    """Object Storage Assets API class.

    Please note that the Object Storage Assets API is currently in beta and
    is subject to change.
    """

    def __init__(self, *args, **kwargs):
        """Inits ObjectStoragesAssets.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__("object-storages", api_version="beta", *args, **kwargs)

    def _format_asset_id(self, asset_id: str) -> str:
        """Formats asset ID.

        Args:
            asset_id (str): Asset ID to format.

        Returns:
            str: Formatted asset ID.
        """
        return quote(asset_id, safe="")



"""Censys ASM CLI."""

import argparse
import concurrent.futures
import csv
import json
import sys
import threading
from typing import Dict, List, Union
from xml.etree import ElementTree

from rich.progress import Progress, TaskID
from rich.prompt import Confirm, Prompt

from censys.asm.api import CensysAsmAPI
from censys.asm.inventory import InventorySearch
from censys.asm.saved_queries import SavedQueries
from censys.asm.seeds import SEED_TYPES, Seeds
from censys.cli.utils import console
from censys.common.config import DEFAULT, get_config, write_config
from censys.common.exceptions import (
    CensysAsmException,
    CensysSeedNotFoundException,
    CensysUnauthorizedException,
)


def cli_asm_config(_: argparse.Namespace):  # pragma: no cover
    """Config asm subcommand.

    Args:
        _: Argparse Namespace.
    """
    api_key_prompt = "Censys ASM API Key"

    config = get_config()
    api_key = config.get(DEFAULT, "asm_api_key")

    if api_key:
        key_len = len(api_key) - 4
        redacted_api_key = api_key.replace(api_key[:key_len], key_len * "*")
        api_key_prompt = f"{api_key_prompt} [cyan]({redacted_api_key})[/cyan]"

    api_key = Prompt.ask(api_key_prompt, console=console) or api_key

    try:
        c = CensysAsmAPI(api_key)
        w = c.get_workspace_id()
        console.print(f"Successfully authenticated to workspace {w}")
    except CensysUnauthorizedException:
        console.print("Failed to authenticate")
        sys.exit(1)

    if not api_key:
        console.print("Please enter valid credentials")
        sys.exit(1)

    color = Confirm.ask(
        "Do you want color output?", default=True, show_default=False, console=console
    )
    config.set(DEFAULT, "color", "auto" if color else "")

    try:
        # Assumes that login was successfully
        config.set(DEFAULT, "asm_api_key", api_key)

        write_config(config)
        console.print("\nSuccessfully configured credentials")
        sys.exit(0)
    except CensysUnauthorizedException:
        console.print("Failed to authenticate")
        sys.exit(1)


def get_seeds_from_xml(file: str) -> List[Dict[str, str]]:
    """Get seeds from nmap xml.

    Args:
        file (str): Nmap xml file.

    Returns:
        List[Dict[str, str]]: List of seeds.
    """
    tree = ElementTree.parse(file)
    root = tree.getroot()
    ips = set()
    domains = set()
    for element in root.findall("host"):
        address_element = element.find("address")
        hostnames_element = element.find("hostnames")
        if address_element is not None and address_element.get("addrtype") == "ipv4":
            ip_address = address_element.get("addr")
            if ip_address:
                ips.add(ip_address)
        if hostnames_element is not None:
            hostname_elements = hostnames_element.findall("hostname")
            for hostname_element in hostname_elements:
                hostname_type = hostname_element.get("type")
                if hostname_type != "user":
                    continue
                hostname = hostname_element.get("name")
                if hostname:
                    domains.add(hostname)
    return [{"value": ip, "type": "IP_ADDRESS"} for ip in ips] + [
        {"value": domain, "type": "DOMAIN_NAME"} for domain in domains
    ]


def get_seeds_from_params(
    args: argparse.Namespace, command_name: str
) -> List[Dict[str, Union[str, int]]]:
    """Get seeds from params.

    Args:
        args (Namespace): Argparse Namespace.
        command_name (str): The name of the command getting the seeds to be processed.

    Returns:
        List[Dict[str, str]]: List of seeds.
    """
    is_csv = args.csv
    if args.input_file or args.json:
        json_data = None
        if args.input_file:
            if args.input_file == "-":
                file = sys.stdin
            else:
                if not args.csv and args.input_file.endswith(".csv"):
                    is_csv = True
                file = open(args.input_file)  # noqa: SIM115

            if is_csv:
                seeds = []
                csv_reader = csv.DictReader(file, delimiter=",")
                if csv_reader.fieldnames:
                    # Lowercase the field names
                    csv_reader.fieldnames = [
                        string.lower() for string in csv_reader.fieldnames
                    ]
                for row in csv_reader:
                    seeds.append(row)
            else:
                json_data = file.read()
        else:
            json_data = args.json

        if json_data:
            try:
                seeds = json.loads(json_data)
            except json.decoder.JSONDecodeError as e:
                console.print(f"Invalid json {e}")
                sys.exit(1)
    elif args.nmap_xml:
        try:
            seeds = get_seeds_from_xml(args.nmap_xml)
        except ElementTree.ParseError as e:
            console.print(f"Invalid xml {e}")
            sys.exit(1)

    seeds_to_add = []
    for seed in seeds:
        if isinstance(seed, str):
            seed = {"value": seed}

        if not isinstance(seed, dict):
            console.print(f"Invalid seed {seed}")
            sys.exit(1)

        if command_name != "delete-seeds" and "type" not in seed:
            seed["type"] = args.default_type

        if (
            command_name != "replace-labeled-seeds"
            and "label" not in seed
            and "label" in args
        ):
            seed["label"] = args.label

        # The back end is really picky about sending extra fields, so we'll prune out anything it won't like.
        # This makes it possible to output to CSV, edit, and then pipe the same CSV back into add-seeds, where
        # we will strip the id, source, createdOn, and any other junk that might be there
        #
        valid_params = ["type", "value"]
        if command_name == "add-seeds":
            valid_params.append("label")
        elif command_name == "delete-seeds" and "value" not in seed:
            valid_params.append("id")
        filtered_seeds = {key: seed[key] for key in seed if key in valid_params}
        seeds_to_add.append(filtered_seeds)

    return seeds_to_add


def console_clear_line():
    """Clear previous line in console."""
    sys.stdout.write("\033[F")  # Move the cursor to the previous line
    sys.stdout.write("\033[K")  # Clear the line


def cli_add_seeds(args: argparse.Namespace):
    """Add seed subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    seeds_to_add = get_seeds_from_params(args, "add-seeds")

    s = Seeds(args.api_key)
    to_add_count = len(seeds_to_add)
    res = s.add_seeds(seeds_to_add)
    added_seeds = res["addedSeeds"]
    added_count = len(added_seeds)
    if not added_count:
        console.print("No seeds were added. (Run with -v to get more info)")
        if not args.verbose:
            sys.exit(1)
    else:
        console.print(f"Added {added_count} seeds.")
    if added_count < to_add_count:
        console.print(f"Seeds not added: {to_add_count - added_count}")
        if args.verbose:  # pragma: no cover
            console.print(
                "The following seed(s) were not able to be added as they already exist or are reserved."
            )
            for seed in seeds_to_add:
                if not any(s for s in added_seeds if seed["value"] == s["value"]):
                    console.print(f"{seed}")


def cli_delete_seeds(args: argparse.Namespace):
    """Delete seeds subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    seeds_to_delete = get_seeds_from_params(args, "delete-seeds")
    s = Seeds(args.api_key)

    # Get all seeds into a dict indexed by value (for later lookups)
    console.print("Getting seeds...")
    seeds = s.get_seeds()
    console_clear_line()

    seeds_dict_indexed_by_value = {}
    for seed in seeds:
        seeds_dict_indexed_by_value[seed["value"]] = seed

    seed_ids_to_delete = []
    seed_ids_not_found = []
    for seed_to_delete in seeds_to_delete:
        if seed_id := seed_to_delete.get("id"):
            seed_ids_to_delete.append(seed_id)
        elif seed_value := seed_to_delete.get("value"):
            # value may not be in current seeds - can't delete
            if seed_value in seeds_dict_indexed_by_value:
                seed_ids_to_delete.append(seeds_dict_indexed_by_value[seed_value]["id"])
            else:
                seed_ids_not_found.append(seed_value)
        else:
            console.print("Error, no seed id or value for seed.")

    if len(seed_ids_to_delete) == 0:
        console.print("No seeds to delete.")
        sys.exit(0)

    seed_ids_deleted = []

    # Create a lock to protect shared variables
    lock = threading.Lock()

    def delete_seed(seed_id: int, progress: Progress, task_id: TaskID):
        try:
            s.delete_seed_by_id(seed_id)
            with lock:
                nonlocal seed_ids_deleted
                seed_ids_deleted.append(seed_id)
        except CensysSeedNotFoundException:  # pragma: no cover
            with lock:
                nonlocal seed_ids_not_found
                seed_ids_not_found.append(seed_id)
        progress.update(task_id, advance=1)

    # Create a rich Progress instance
    with Progress() as progress:
        progress_task_id = progress.add_task(
            "[cyan]Deleting[/cyan]", total=len(seeds_to_delete)
        )
        tasks = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=25) as executor:
            # Submit requests using the executor
            for seed_id in seed_ids_to_delete:
                if isinstance(seed_id, str):  # pragma: no cover
                    console.print(f"Invalid seed id {seed_id}")
                    continue
                task = executor.submit(delete_seed, seed_id, progress, progress_task_id)
                tasks.append(task)

            # Wait for all requests to complete
            concurrent.futures.wait(tasks)

    console.print(f"Deleted {len(seed_ids_deleted)} seeds.")
    if len(seed_ids_not_found) > 0:
        console.print(
            f"Unable to delete {len(seed_ids_not_found)} seeds because they were not present."
        )


def cli_delete_all_seeds(args: argparse.Namespace):
    """Delete all seeds subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    if not args.force:
        delete_all = Confirm.ask(
            "Are you sure you want to delete all seeds?",
            default=False,
            show_default=True,
            console=console,
        )
        console_clear_line()
        if not delete_all:
            sys.exit(1)

    s = Seeds(args.api_key)

    console.print("Getting seeds...")
    seeds = s.get_seeds()
    console_clear_line()

    # Create a lock to protect shared variables
    lock = threading.Lock()
    seeds_delete_count = 0

    def delete_seed(seed_id: int, progress: Progress, task_id: TaskID):
        s.delete_seed_by_id(seed_id)
        with lock:
            nonlocal seeds_delete_count
            seeds_delete_count += 1
        progress.update(task_id, advance=1)

    # Create a rich Progress instance
    with Progress() as progress:
        progress_task_id = progress.add_task("[cyan]Deleting[/cyan]", total=len(seeds))
        tasks = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=25) as executor:
            # Submit requests using the executor
            for seed in seeds:
                task = executor.submit(
                    delete_seed, seed["id"], progress, progress_task_id
                )
                tasks.append(task)

            # Wait for all requests to complete
            concurrent.futures.wait(tasks)

    console.print(f"Deleted {seeds_delete_count} of {len(seeds)} total seeds.")


def cli_delete_seeds_with_label(args: argparse.Namespace):
    """Delete seeds with label subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    s = Seeds(args.api_key)
    s.delete_seeds_by_label(args.label)
    console.print(f'Deleted seeds with label "{args.label}".')


def cli_replace_seeds_with_label(args: argparse.Namespace):
    """Replace seeds with label subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    seeds_to_add = get_seeds_from_params(args, "replace-labeled-seeds")
    s = Seeds(args.api_key)

    res = s.replace_seeds_by_label(args.label, seeds_to_add, True)
    console.print(
        f"Removed {len(res['removedSeeds'])} seeds.  Added {len(res['addedSeeds'])} seeds.  Skipped {len(res['skippedReservedSeeds'])} reserved seeds."
    )

    if args.verbose:  # pragma: no cover
        added_seeds = res.get("addedSeeds", [])
        if len(added_seeds) > 0:
            console.print("The following seed(s) were added.")
            for added_seed in added_seeds:
                console.print(f"    {added_seed}")
        removed_seeds = res.get("removedSeeds", [])
        if len(removed_seeds) > 0:
            console.print("The following seed(s) were removed.")
            for removed_seed in removed_seeds:
                console.print(f"    {removed_seed}")
        skipped_seeds = res.get("skippedReservedSeeds", [])
        if len(skipped_seeds) > 0:
            console.print(
                "The following seed(s) were not added because they are reserved."
            )
            for skipped_seed in skipped_seeds:
                console.print(f"    {skipped_seed}")


def cli_list_seeds(args: argparse.Namespace):
    """List seeds subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    s = Seeds(args.api_key)
    res = s.get_seeds(args.type, args.label)
    if args.csv:
        console.print("id,type,value,label,source,createdOn")
        for seed in res:
            console.print(
                f"{seed['id']},{seed['type']},{seed['value']},{seed['label']},{seed['source']},{seed['createdOn']}"  # noqa: E231
            )
    else:
        console.print_json(json.dumps(res))


def add_seed_arguments(parser: argparse._SubParsersAction, is_delete=False) -> None:
    """Add seed arguments to parser.

    Args:
        parser (argparse._SubParsersAction): Parser.
        is_delete (Boolean): Is this a delete command.
    """
    if not is_delete:
        parser.add_argument(  # type: ignore[attr-defined]
            "--default-type",
            help="type of the seed(s) if type is not already provided (default: %(default)s)",
            choices=SEED_TYPES,
            default="IP_ADDRESS",
        )
    parser.add_argument("--csv", help="process input in CSV format", action="store_true")  # type: ignore[attr-defined]
    seeds_group = parser.add_mutually_exclusive_group(required=True)  # type: ignore[attr-defined]
    seeds_group.add_argument(
        "--input-file",
        "-i",
        help="input file name containing valid seeds in JSON format, unless --csv is specified (use - for stdin)",
        type=str,
    )
    seeds_group.add_argument(
        "--json", "-j", help="input string containing valid json seeds", type=str
    )
    seeds_group.add_argument(
        "--nmap-xml", help="input file name containing valid xml nmap output", type=str
    )


def cli_list_saved_queries(args: argparse.Namespace):
    """List saved queries subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    s = SavedQueries(args.api_key)
    try:
        res = s.get_saved_queries(
            args.query_name_prefix, args.page_size, args.page, args.filter_term
        )

        if args.csv:
            console.print("queryId,queryName,query,createdAt")
            for query in res["results"]:
                console.print(
                    f"{query['queryId']},{query['queryName']},{query['query']},{query['createdAt']}"  # noqa: E231
                )
        else:
            console.print_json(json.dumps(res))
    except KeyError:
        console.print("Failed to get saved queries.")
        sys.exit(1)


def cli_add_saved_query(args: argparse.Namespace):
    """Add saved query subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    s = SavedQueries(args.api_key)
    res = s.add_saved_query(args.query, args.query_name)
    try:
        console.print(
            f"Added saved query.\nQuery name: {res['result']['queryName']}\nQuery: {res['result']['query']}\nQuery ID: {res['result']['queryId']}\nCreated at: {res['result']['createdAt']}"
        )
    except (KeyError, CensysAsmException):
        console.print("Failed to add saved query.")
        sys.exit(1)


def cli_get_saved_query_by_id(args: argparse.Namespace):
    """Get saved query by id subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    s = SavedQueries(args.api_key)
    res = s.get_saved_query_by_id(args.query_id)
    try:
        console.print(
            f"Query name: {res['result']['queryName']}\nQuery: {res['result']['query']}\nQuery ID: {res['result']['queryId']}\nCreated at: {res['result']['createdAt']}"
        )
    except (KeyError, CensysAsmException):
        console.print("Failed to get saved query.")
        sys.exit(1)


def cli_edit_saved_query_by_id(args: argparse.Namespace):
    """Edit saved query by id subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    s = SavedQueries(args.api_key)
    res = s.edit_saved_query_by_id(args.query_id, args.query, args.query_name)
    try:
        console.print(
            f"Edited saved query.\nQuery name: {res['result']['queryName']}\nQuery: {res['result']['query']}\nQuery ID: {res['result']['queryId']}\nCreated at: {res['result']['createdAt']}"
        )
    except (KeyError, CensysAsmException):
        console.print("Failed to edit saved query.")
        sys.exit(1)


def cli_delete_saved_query_by_id(args: argparse.Namespace):
    """Delete saved query by id subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    s = SavedQueries(args.api_key)
    res = s.delete_saved_query_by_id(args.query_id)
    if res == {}:
        console.print(f"Deleted saved query with ID {args.query_id}.")
    else:
        console.print("Failed to delete saved query.")
        sys.exit(1)


def cli_execute_saved_query_by_name(args: argparse.Namespace):
    """Execute saved query by name subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    # do some sanity checking on page size before anything else
    if args.page_size > 1000:
        console.print(
            "page size must be within [0,1000]. To fetch all pages, specify --pages -1 with any legal page size"
        )

        sys.exit(1)
    s = InventorySearch(args.api_key)
    q = SavedQueries(args.api_key)
    # get query from name
    queries = q.get_saved_queries(args.query_name, 1, 1)
    results = queries.get("results")
    if not results:
        console.print("No saved query found with that name.")
        sys.exit(1)
    query = results[0]["query"]

    try:
        res = s.search(
            None, query, args.page_size, None, args.sort, args.fields, args.pages
        )
        console.print_json(json.dumps(res))
    except CensysAsmException:
        console.print("Failed to execute saved query.")
        sys.exit(1)


def cli_execute_saved_query_by_id(args: argparse.Namespace):
    """Execute saved query by id subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    # do some sanity checking on page size before anything else
    if args.page_size > 1000:
        console.print(
            "page size must be within [0,1000]. To fetch all pages, specify --pages -1 with any legal page size"
        )

        sys.exit(1)
    s = InventorySearch(args.api_key)
    q = SavedQueries(args.api_key)
    try:
        query_json = q.get_saved_query_by_id(args.query_id)
        query = query_json["result"]["query"]
    except (KeyError, CensysAsmException):
        console.print("No saved query found with that ID.")
        sys.exit(1)
    try:
        res = s.search(
            None, query, args.page_size, None, args.sort, args.fields, args.pages
        )
        console.print_json(json.dumps(res))
    except CensysAsmException:
        console.print("Failed to execute saved query.")
        sys.exit(1)


def cli_search(args: argparse.Namespace):
    """Inventory search subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    s = InventorySearch(args.api_key)

    try:
        res = s.search(
            args.workspaces,
            args.query,
            args.page_size,
            args.cursor,
            args.sort,
            args.fields,
            args.pages,
        )
        console.print_json(json.dumps(res))
    except CensysAsmException:
        console.print("Failed to execute query.")
        sys.exit(1)


def include(parent_parser: argparse._SubParsersAction, parents: dict):
    """Include this subcommand into the parent parser.

    Args:
        parent_parser (argparse._SubParsersAction): Parent parser.
        parents (dict): Parent arg parsers.
    """
    asm_parser = parent_parser.add_parser(
        "asm", description="Interact with the Censys ASM API", help="interact with ASM"
    )

    def add_verbose(parser):
        parser.add_argument(
            "-v",
            "--verbose",
            help="verbose output",
            action="store_true",
        )

    asm_subparser = asm_parser.add_subparsers()

    # Add config command
    config_parser = asm_subparser.add_parser(
        "config",
        description="Configure Censys ASM API Settings",
        help="configure Censys ASM API settings",
    )
    config_parser.set_defaults(func=cli_asm_config)

    add_parser = asm_subparser.add_parser(
        "add-seeds",
        description="Add seeds to ASM",
        help="add seeds",
        parents=[parents["asm_auth"]],
    )
    add_verbose(add_parser)
    add_seed_arguments(add_parser)
    add_parser.add_argument(
        "-l",
        "--label",
        help='label to apply to seeds without label (default: "")',
        type=str,
        default="",
    )
    add_parser.set_defaults(func=cli_add_seeds)

    delete_parser = asm_subparser.add_parser(
        "delete-seeds",
        description="Delete ASM seeds",
        help="delete seeds",
        parents=[parents["asm_auth"]],
    )
    add_verbose(delete_parser)
    add_seed_arguments(delete_parser, True)
    delete_parser.set_defaults(func=cli_delete_seeds)

    delete_all_parser = asm_subparser.add_parser(
        "delete-all-seeds",
        description="Delete all ASM seeds",
        help="delete all seeds",
        parents=[parents["asm_auth"]],
    )
    delete_all_parser.add_argument(
        "-f",
        "--force",
        help="force delete all (no confirmation prompt)",
        action="store_true",
    )
    add_verbose(delete_all_parser)
    delete_all_parser.set_defaults(func=cli_delete_all_seeds)

    delete_with_label_parser = asm_subparser.add_parser(
        "delete-labeled-seeds",
        description="Delete all ASM seeds with specified label",
        help="delete all seeds having label",
        parents=[parents["asm_auth"]],
    )
    delete_with_label_parser.add_argument(
        "-l",
        "--label",
        help="label for which to delete all seeds",
        type=str,
        default="",
        required=True,
    )
    add_verbose(delete_with_label_parser)
    delete_with_label_parser.set_defaults(func=cli_delete_seeds_with_label)

    replace_with_label_parser = asm_subparser.add_parser(
        "replace-labeled-seeds",
        description="Replace all ASM seeds with specified label with new seeds",
        help="replace all seeds having label",
        parents=[parents["asm_auth"]],
    )
    replace_with_label_parser.add_argument(
        "-l",
        "--label",
        help="label for which to replace all seeds",
        type=str,
        default="",
        required=True,
    )
    add_verbose(replace_with_label_parser)
    add_seed_arguments(replace_with_label_parser)
    replace_with_label_parser.set_defaults(func=cli_replace_seeds_with_label)

    list_parser = asm_subparser.add_parser(
        "list-seeds",
        description="List all ASM seeds, optionally filtered by label and type",
        help="list seeds",
        parents=[parents["asm_auth"]],
    )
    list_parser.add_argument(
        "-t",
        "--type",
        help="type of the seed to list, if not specified, all types returned",
        choices=SEED_TYPES,
        default="",
    )
    list_parser.add_argument(
        "-l",
        "--label",
        help="label of seeds to list, if not specified, all labels returned",
        type=str,
        default="",
    )
    list_parser.add_argument(
        "--csv", help="output in CSV format (otherwise JSON)", action="store_true"
    )
    add_verbose(list_parser)
    list_parser.set_defaults(func=cli_list_seeds)

    list_saved_queries_parser = asm_subparser.add_parser(
        "list-saved-queries",
        description="List all ASM saved queries, optionally filtered by query name prefix and filter term",
        help="list saved queries",
        parents=[parents["asm_auth"]],
    )
    list_saved_queries_parser.add_argument(
        "--query-name-prefix",
        help="Prefix for the saved query name to filter by",
        type=str,
        default="",
    )
    list_saved_queries_parser.add_argument(
        "--filter-term",
        help="Term used to filter the list of saved query names and the saved queries",
        type=str,
        default="",
    )
    list_saved_queries_parser.add_argument(
        "--page-size",
        help="Number of results to return. Defaults to 50.",
        type=int,
        default=50,
    )
    list_saved_queries_parser.add_argument(
        "--page",
        help="Page number to begin at when searching. Defaults to 1.",
        type=int,
        default=1,
    )
    list_saved_queries_parser.add_argument(
        "--csv", help="output in CSV format (otherwise JSON)", action="store_true"
    )
    add_verbose(list_saved_queries_parser)
    list_saved_queries_parser.set_defaults(func=cli_list_saved_queries)

    add_saved_query_parser = asm_subparser.add_parser(
        "add-saved-query",
        description="Add a saved query to ASM",
        help="add saved query",
        parents=[parents["asm_auth"]],
    )
    add_saved_query_parser.add_argument(
        "--query-name",
        help="Name of the saved query",
        type=str,
        required=True,
    )
    add_saved_query_parser.add_argument(
        "--query",
        help="Query string",
        type=str,
        required=True,
    )
    add_verbose(add_saved_query_parser)
    add_saved_query_parser.set_defaults(func=cli_add_saved_query)

    get_saved_query_by_id_parser = asm_subparser.add_parser(
        "get-saved-query-by-id",
        description="Get a saved query by ID",
        help="get saved query by ID",
        parents=[parents["asm_auth"]],
    )
    get_saved_query_by_id_parser.add_argument(
        "--query-id",
        help="ID of the saved query",
        type=str,
        required=True,
    )
    add_verbose(get_saved_query_by_id_parser)
    get_saved_query_by_id_parser.set_defaults(func=cli_get_saved_query_by_id)

    edit_saved_query_by_id_parser = asm_subparser.add_parser(
        "edit-saved-query-by-id",
        description="Edit a saved query by ID",
        help="edit saved query by ID",
        parents=[parents["asm_auth"]],
    )
    edit_saved_query_by_id_parser.add_argument(
        "--query-id",
        help="ID of the saved query",
        type=str,
        required=True,
    )
    edit_saved_query_by_id_parser.add_argument(
        "--query-name",
        help="Name of the saved query",
        type=str,
        required=True,
    )
    edit_saved_query_by_id_parser.add_argument(
        "--query",
        help="Query string",
        type=str,
        required=True,
    )
    add_verbose(edit_saved_query_by_id_parser)
    edit_saved_query_by_id_parser.set_defaults(func=cli_edit_saved_query_by_id)

    delete_saved_query_by_id_parser = asm_subparser.add_parser(
        "delete-saved-query-by-id",
        description="Delete a saved query by ID",
        help="delete saved query by ID",
        parents=[parents["asm_auth"]],
    )
    delete_saved_query_by_id_parser.add_argument(
        "--query-id",
        help="ID of the saved query",
        type=str,
        required=True,
    )
    add_verbose(delete_saved_query_by_id_parser)
    delete_saved_query_by_id_parser.set_defaults(func=cli_delete_saved_query_by_id)

    execute_saved_query_by_name_parser = asm_subparser.add_parser(
        "execute-saved-query-by-name",
        description="Execute a saved query by name in inventory search",
        help="execute saved query by name",
        parents=[parents["asm_auth"]],
    )
    execute_saved_query_by_name_parser.add_argument(
        "--query-name",
        help="Query name",
        type=str,
        required=True,
    )
    execute_saved_query_by_name_parser.add_argument(
        "--page-size",
        help="Number of results to return. Defaults to 50.",
        type=int,
        default=50,
    )
    execute_saved_query_by_name_parser.add_argument(
        "--sort",
        help="Sort order for results",
        type=List[str],
        default=[],
    )
    execute_saved_query_by_name_parser.add_argument(
        "--fields",
        help="Fields to include in results",
        type=List[str],
        default=[],
    )
    execute_saved_query_by_name_parser.add_argument(
        "--pages",
        help="Number of pages to return. Defaults to 1.",
        type=int,
        default=1,
    )
    add_verbose(execute_saved_query_by_name_parser)
    execute_saved_query_by_name_parser.set_defaults(
        func=cli_execute_saved_query_by_name
    )

    execute_saved_query_by_id_parser = asm_subparser.add_parser(
        "execute-saved-query-by-id",
        description="Execute a saved query by id in inventory search",
        help="execute saved query by id",
        parents=[parents["asm_auth"]],
    )
    execute_saved_query_by_id_parser.add_argument(
        "--query-id",
        help="Query ID",
        type=str,
        required=True,
    )
    execute_saved_query_by_id_parser.add_argument(
        "--page-size",
        help="Number of results to return. Defaults to 50.",
        type=int,
        default=50,
    )
    execute_saved_query_by_id_parser.add_argument(
        "--sort",
        help="Sort order for results",
        type=List[str],
        default=[],
    )
    execute_saved_query_by_id_parser.add_argument(
        "--fields",
        help="Fields to include in results",
        type=List[str],
        default=[],
    )
    execute_saved_query_by_id_parser.add_argument(
        "--pages",
        help="Number of pages to return. Defaults to 1.",
        type=int,
        default=1,
    )
    add_verbose(execute_saved_query_by_id_parser)
    execute_saved_query_by_id_parser.set_defaults(func=cli_execute_saved_query_by_id)

    search_parser = asm_subparser.add_parser(
        "search",
        description="Execute a query in inventory search",
        help="execute query in inventory search",
        parents=[parents["asm_auth"]],
    )
    search_parser.add_argument(
        "--query",
        help="Query string",
        type=str,
        required=True,
    )
    search_parser.add_argument(
        "--page-size",
        help="Number of results to return. Defaults to 50.",
        type=int,
        default=50,
    )
    search_parser.add_argument(
        "--cursor",
        help="Cursor to use for pagination",
        type=str,
        default="",
    )
    search_parser.add_argument(
        "--sort",
        help="Sort order for results",
        type=List[str],
        default=[],
    )
    search_parser.add_argument(
        "--fields",
        help="Fields to include in results",
        type=List[str],
        default=[],
    )
    search_parser.add_argument(
        "--workspaces",
        help="Workspace IDs to search. Deprecated. The workspace associated with `CENSYS-API-KEY` will be used automatically.",
        type=str,
        required=False,
    )
    search_parser.add_argument(
        "--pages",
        help="Number of pages to return. Defaults to 1.",
        type=int,
        default=1,
    )
    add_verbose(search_parser)
    search_parser.set_defaults(func=cli_search)



"""Censys CLI utilities."""

import argparse
import datetime
import json
import os.path
import sys
from typing import Any, Dict, List, Optional, Union

from rich.console import Console

from censys.common.config import DEFAULT, get_config

Results = Union[List[dict], Dict[str, Any]]

V2_INDEXES = ["hosts", "certificates"]

config = get_config()
color = config.get(DEFAULT, "color")
color_system = "auto" if color else None
console = Console(color_system=color_system)  # type: ignore
err_console = Console(color_system=color_system, file=sys.stderr)  # type: ignore


def print_wrote_file(file_path: str):
    """Print wrote file confirmation.

    Args:
        file_path (str): Name of the file to write to on the disk.
    """
    abs_file_path = os.path.abspath(file_path)
    console.print(f"Wrote results to file {abs_file_path}", soft_wrap=True)


def _write_json(file_path: str, search_results: Results):
    """Write search results to a new file in JSON format.

    Args:
        file_path (str): Name of the file to write to on the disk.
        search_results (Results): A list of results from the query.
    """
    with open(file_path, "w") as output_file:
        # Since the results are already in JSON, just write them to a file.
        json.dump(search_results, output_file, indent=4)

    print_wrote_file(file_path)


def _write_screen(search_results: Results):  # pragma: no cover
    """Writes search results to standard output.

    Args:
        search_results (Results): A list of results from the query.
    """
    config = get_config()
    if config.get(DEFAULT, "color"):
        console.print_json(data=search_results)
    else:
        print(json.dumps(search_results, indent=4))


def write_file(
    results_list: Results,
    file_format: Optional[str] = None,
    file_path: Optional[str] = None,
    csv_fields: Optional[List[str]] = None,
):
    """Maps formats and writes results.

    Args:
        results_list (Results): A list of results from the API query.
        file_format (str): Optional; The format of the output.
        file_path (str): Optional; A path to write results to.
        csv_fields (List[str]): Optional; A list of fields to write to CSV.
    """
    if file_format and isinstance(file_format, str):
        file_format = file_format.lower()

    if not file_path:
        file_path = "temp-out.json"

    if file_format == "json":
        _write_json(file_path, results_list)
    else:
        _write_screen(results_list)


def valid_datetime_type(datetime_str: str) -> datetime.datetime:
    """Custom argparse type for user datetime values from arg.

    Args:
        datetime_str (str): A string representing a datetime.

    Raises:
        ArgumentTypeError: If the datetime string is invalid.

    Returns:
        datetime.datetime: A datetime object.
    """
    try:
        return datetime.datetime.strptime(datetime_str, "%Y-%m-%d %H:%M")
    except ValueError:
        try:
            return datetime.datetime.strptime(datetime_str, "%Y-%m-%d")
        except ValueError:
            msg = f"Given datetime ({datetime_str}) is not valid! Expected format: 'YYYY-MM-DD' or 'YYYY-MM-DD HH:mm'."  # noqa: E231
            raise argparse.ArgumentTypeError(msg)



"""Interact with miscellaneous Censys Beta APIs."""

from typing import List, Optional

from ..common.types import Datetime
from ..common.utils import format_iso8601
from .api import CensysAsmAPI


class Beta(CensysAsmAPI):
    """Beta API class."""

    base_path = "/beta"

    def get_logbook_data(
        self, filters: Optional[dict] = None, cursor: Optional[str] = None
    ):
        """Retrieve logbook data.

        Args:
            filters (dict): Optional; Filter parameters.
            cursor (str): Optional; Cursor for pagination.

        Returns:
            dict: Logbook data result.
        """
        data = {"filters": filters, "nextWindowCursor": cursor}
        return self._post(f"{self.base_path}/logbook/getLogbookData", data=data)

    def add_cloud_assets(
        self,
        cloud_connector_uid: str,
        cloud_assets: List[dict],
    ):
        """Add cloud assets.

        Args:
            cloud_connector_uid (str): Cloud connector UID.
            cloud_assets (List[dict]): Cloud assets.

        Returns:
            dict: Add cloud assets result.
        """
        data = {
            "cloudConnectorUid": cloud_connector_uid,
            "cloudAssets": cloud_assets,
        }
        return self._post(f"{self.base_path}/cloudConnector/addCloudAssets", data=data)

    def get_input_assets(self, page_number: int = 1, page_size: Optional[int] = None):
        """Retrieve input assets.

        Args:
            page_number (int): Optional; Page number to begin at when searching.
            page_size (int): Optional; Page size for retrieving assets.

        Returns:
            dict: Input assets result.
        """
        return self._get(
            f"{self.base_path}/assets/inputAssets",
            params={"pageNumber": page_number, "pageSize": page_size},
        )

    def get_asset_counts(self, since: Datetime, environment: str, asset_type: str):
        """Retrieve asset counts.

        Args:
            since (Datetime): Date to include assets from.
            environment (str): Environment to include assets from.
            asset_type (str): Asset type to include.

        Returns:
            dict: Asset count result.
        """
        params = {
            "since": format_iso8601(since),
            "environment": environment,
            "assetType": asset_type,
        }
        return self._get(
            f"{self.base_path}/assets/counts",
            params=params,
        )

    def get_host_counts_by_country(self, since: Datetime, environment: str):
        """Retrieve host counts by country.

        Args:
            since (Datetime): Date to include hosts from.
            environment (str): Environment to include hosts from.

        Returns:
            dict: Host count result.
        """
        params = {
            "since": format_iso8601(since),
            "environment": environment,
        }
        return self._get(
            f"{self.base_path}/assets/hostCountsByCountry",
            params=params,
        )

    def get_user_workspaces(self, user_uuid: str):
        """Retrieve user workspaces.

        Args:
            user_uuid (str): User UUID.

        Returns:
            dict: User workspaces result.
        """
        return self._get(
            f"{self.base_path}/users/{user_uuid}/workspaces",
        )



"""Interact with the Censys Saved Queries API."""

from typing import Optional

from .api import CensysAsmAPI


class SavedQueries(CensysAsmAPI):
    """Saved Queries API class."""

    base_path = "/inventory/v1/saved-query"

    def get_saved_queries(
        self,
        query_name_prefix: Optional[str] = None,
        page_size: Optional[int] = None,
        page: Optional[int] = None,
        filter_term: Optional[str] = None,
    ) -> dict:
        """Get saved queries.

        Args:
            query_name_prefix (str, optional): Prefix for the saved query name.
            page_size (int, optional): Number of results to return. Defaults to 50.
            page (int, optional): Page number to begin at when searching. Defaults to 1.
            filter_term (str, optional): Term used to filter the list of saved query names and the saved queries.

        Returns:
            dict: Saved queries results.
        """
        if page_size is None:
            page_size = 50
        if page is None:
            page = 1
        args: dict = {
            "pageSize": page_size,
            "page": page,
        }

        if query_name_prefix:
            args["queryNamePrefix"] = query_name_prefix
        if filter_term:
            args["filterTerm"] = filter_term

        return self._get(self.base_path, args=args)

    def add_saved_query(
        self,
        query: str,
        query_name: str,
    ) -> dict:
        """Add a new saved query to the ASM platform.

        Args:
            query (str): Query string.
            query_name (str): Saved query name.

        Returns:
            dict: Added saved query results.
        """
        body = {
            "query": query,
            "queryName": query_name,
        }

        return self._post(self.base_path, data=body)

    def get_saved_query_by_id(
        self,
        query_id: str,
    ) -> dict:
        """Get saved query by query ID.

        Args:
            query_id (str): The saved query's ID.

        Returns:
            dict: Saved query result.
        """
        return self._get(f"{self.base_path}/{query_id}")

    def edit_saved_query_by_id(
        self,
        query_id: str,
        query: str,
        query_name: str,
    ) -> dict:
        """Edit an existing saved query by query ID.

        Args:
            query_id (str): The saved query's ID.
            query (str): New query string.
            query_name (str): New saved query name.

        Returns:
            dict: Edited saved query result.
        """
        body = {
            "query": query,
            "queryName": query_name,
        }

        return self._put(f"{self.base_path}/{query_id}", data=body)

    def delete_saved_query_by_id(
        self,
        query_id: str,
    ) -> dict:
        """Delete saved query by query ID.

        Args:
            query_id (str): The saved query's ID.

        Returns:
            dict: Delete results.
        """
        return self._delete(f"{self.base_path}/{query_id}")



"""Interact with the Censys Risks API."""

import urllib.parse
from typing import Any, Dict, List, Optional

from .api import CensysAsmAPI


class Risks(CensysAsmAPI):
    """Risks API class."""

    base_path = "/v2/risk"
    risk_events_path = f"{base_path}-events"
    risk_instances_path = f"{base_path}-instances"
    risk_types_path = f"{base_path}-types"

    def get_risk_events(
        self,
        start: Optional[str] = None,
        end: Optional[str] = None,
        after_id: Optional[int] = None,
        limit: Optional[int] = None,
        cursor: Optional[str] = None,
        accept: Optional[str] = None,
    ) -> dict:
        """Retrieve risk events.

        Args:
            start (str): Optional; Starting event time, inclusive (in RFC3339 format).
            end (str): Optional; Ending event time, inclusive (in RFC3339 format).
            after_id (int): Optional; Risk event ID to query for events after.
            limit (int): Optional; Max number of events to return.
            cursor (str): Optional; Cursor value to continue collecting events started in a previous request.
            accept (str): Optional; Accept header.

        Returns:
            dict: Risk events result.
        """
        args: Dict[str, Any] = {}
        if start:
            args["start"] = start
        if end:
            args["end"] = end
        if after_id:
            args["afterID"] = after_id
        if limit:
            args["limit"] = limit
        if cursor:
            args["cursor"] = cursor
        return self._get(
            self.risk_events_path,
            args=args,
            headers={"Accept": accept} if accept else None,
        )

    def get_risk_instances(
        self, include_events: Optional[bool] = None, accept: Optional[str] = None
    ) -> dict:
        """Retrieve risk instances.

        Args:
            include_events (bool): Optional; Whether to include events.
            accept (str): Optional; Accept header.

        Returns:
            dict: Risk instances result.
        """
        args = {"includeEvents": include_events}
        return self._get(
            self.risk_instances_path,
            args=args,
            headers={"Accept": accept} if accept else None,
        )

    def patch_risk_instances(self, data: dict) -> dict:
        """Patch risk instances.

        Args:
            data (dict): Risk instances data.

        Returns:
            dict: Risk instances result.
        """
        return self._patch(self.risk_instances_path, data=data)

    def search_risk_instances(self, data: dict, accept: Optional[str] = None) -> dict:
        """Search risk instances.

        Args:
            data (dict): Query data.
            accept (str): Optional; Accept header.

        Returns:
            dict: Risk instances result.
        """
        return self._post(
            f"{self.risk_instances_path}/search",
            data=data,
            headers={"Accept": accept} if accept else None,
        )

    def get_risk_instance(
        self, risk_instance_id: int, include_events: Optional[bool] = None
    ) -> dict:
        """Retrieve a risk instance.

        Args:
            risk_instance_id (int): Risk instance ID.
            include_events (bool): Optional; Whether to include events.

        Returns:
            dict: Risk instance result.
        """
        args = {"includeEvents": include_events}
        return self._get(f"{self.risk_instances_path}/{risk_instance_id}", args=args)

    def patch_risk_instance(self, risk_instance_id: int, data: dict) -> dict:
        """Patch a risk instance.

        Args:
            risk_instance_id (int): Risk instance ID.
            data (dict): Risk instance data.

        Returns:
            dict: Risk instance result.
        """
        return self._patch(f"{self.risk_instances_path}/{risk_instance_id}", data=data)

    def get_risk_types(
        self,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        sort: Optional[List[str]] = None,
        include_events: Optional[bool] = None,
        accept: Optional[str] = None,
    ) -> dict:
        """Retrieve risk types.

        Args:
            limit (int, optional): Maximum number of results to return. Defaults to 1000.
            page (int, optional): Page number to begin at when searching. Defaults to 1.
            sort (list): Optional; Sort by field(s).
            include_events (bool): Optional; Whether to include events.
            accept (str): Optional; Accept header.

        Returns:
            dict: Risk types result.
        """
        args: Dict[str, Any] = {"sort": sort, "includeEvents": include_events}
        if page:
            args["page"] = page
        if limit:
            args["limit"] = limit
        return self._get(
            self.risk_types_path,
            args=args,
            headers={"Accept": accept} if accept else None,
        )

    def get_risk_type(
        self, risk_type: str, include_events: Optional[bool] = None
    ) -> dict:
        """Retrieve a risk type.

        Args:
            risk_type (str): Risk type.
            include_events (bool): Optional; Whether to include events.

        Returns:
            dict: Risk type result.
        """
        escaped_risk_type = urllib.parse.quote(risk_type, safe="")
        args = {"includeEvents": include_events}
        return self._get(f"{self.risk_types_path}/{escaped_risk_type}", args=args)

    def patch_risk_type(self, risk_type: str, data: dict) -> dict:
        """Patch a risk type.

        Args:
            risk_type (str): Risk type.
            data (dict): Risk type data.

        Returns:
            dict: Risk type result.
        """
        escaped_risk_type = urllib.parse.quote(risk_type, safe="")
        return self._patch(f"{self.risk_types_path}/{escaped_risk_type}", data=data)



"""Interact with the Censys Inventory Search API."""

import warnings
from typing import List, Optional

from .api import CensysAsmAPI


class InventorySearch(CensysAsmAPI):
    """Inventory Search API class."""

    base_path = "/inventory/v1"

    def search(
        self,
        workspaces: Optional[List[str]] = None,
        query: Optional[str] = None,
        page_size: Optional[int] = None,
        cursor: Optional[str] = None,
        sort: Optional[List[str]] = None,
        fields: Optional[List[str]] = None,
        pages: Optional[int] = None,
    ) -> dict:
        """Search inventory data.

        Args:
            workspaces (List[str], optional): List of workspace IDs to search. Deprecated. The workspace associated with `CENSYS-API-KEY` will be used automatically.
            query (str, optional): Query string.
            page_size (int, optional): Number of results to return. Defaults to 50.
            cursor (str, optional): Cursor to start search from.
            sort (List[str], optional): List of fields to sort by.
            fields (List[str], optional): List of fields to return.
            pages (int, optional): Number of pages of results to return (when set to -1 returns all pages available).

        Returns:
            dict: Inventory search results.
        """
        if workspaces is None:
            workspaces = [self.get_workspace_id()]
        else:
            warnings.warn(
                "The field 'workspaces' is being deprecated. The workspace associated with `CENSYS-API-KEY` will be used automatically.",
                category=DeprecationWarning,
                stacklevel=2,
            )
        if page_size is None:
            page_size = 50

        if pages is None:
            pages = 1

        args = {
            "workspaces": workspaces,
            "pageSize": page_size,
        }

        if query:
            args["query"] = query
        if cursor:
            args["cursor"] = cursor
        if sort:
            args["sort"] = sort
        if fields:
            args["fields"] = fields

        page = 1
        next_cursor = None
        hits = []
        resp = self._get(self.base_path, args=args)
        next_cursor = resp.get("nextCursor")
        hits.extend(resp.get("hits", []))
        # Fetch additional pages if next_cursor is available AND additional pages are requested
        # Loop will exit if next_cursor is None or if the number of pages requested is reached
        # Loop will exit if non-200 status code is returned
        while next_cursor and (pages == -1 or page < pages):
            args["cursor"] = next_cursor
            resp = self._get(self.base_path, args=args)
            if "nextCursor" in resp:
                next_cursor = resp.get("nextCursor")
            else:
                next_cursor = None
            hits.extend(resp.get("hits", []))
            page += 1

        resp["hits"] = hits
        return resp

    def aggregate(
        self,
        workspaces: List[str],
        query: Optional[str] = None,
        aggregation: Optional[dict] = None,
    ) -> dict:
        """Aggregate inventory data.

        Args:
            workspaces (List[str]): List of workspace IDs to search.
            query (str, optional): Query string.
            aggregation (dict, optional): Aggregation object.

        Returns:
            dict: Inventory aggregation results.
        """
        body = {
            "workspaces": workspaces,
            "query": query,
            "aggregation": aggregation,
        }

        return self._post(f"{self.base_path}/aggregate", data=body)

    def fields(self, fields: Optional[List[str]] = None) -> dict:
        """List inventory fields.

        If no fields are specified, all fields will be returned.

        Args:
            fields (List[str], optional): List of fields to return.

        Returns:
            dict: Inventory field results.
        """
        args = {"fields": fields}

        return self._get(f"{self.base_path}/fields", args=args)


Inventory = InventorySearch



"""Censys subdomains CLI."""

import argparse
import json
import sys
from typing import List, Set

from censys.cli.utils import console, err_console
from censys.common.exceptions import (
    CensysException,
    CensysRateLimitExceededException,
    CensysUnauthorizedException,
)
from censys.search import CensysCerts


def print_subdomains(subdomains: Set[str], as_json: bool = False):
    """Print subdomains.

    Args:
        subdomains (Set[str]): Subdomains.
        as_json (bool): Output in JSON format.
    """
    if as_json:
        console.print_json(json.dumps(list(subdomains)))
    else:
        for subdomain in subdomains:
            console.print(f"  - {subdomain}")  # noqa: E221


def cli_subdomains(args: argparse.Namespace):  # pragma: no cover
    """Subdomain subcommand.

    Args:
        args: Argparse Namespace.
    """
    subdomains = set()
    try:
        client = CensysCerts(api_id=args.api_id, api_secret=args.api_secret)
        certificate_query = f"names: {args.domain}"

        with err_console.status(f"Querying {args.domain} subdomains"):
            query = client.search(
                certificate_query, per_page=100, pages=args.pages
            )  # 100 is the max per page

            # Flatten the result, and remove duplicates
            for hits in query:
                for cert in hits:
                    new_subdomains: List[str] = cert.get("names", [])
                    subdomains.update(
                        [
                            subdomain
                            for subdomain in new_subdomains
                            if subdomain.endswith(args.domain)
                        ]
                    )

        # Don't make console prints if we're in json mode
        if not args.json:
            if len(subdomains) == 0:
                err_console.print(f"No subdomains found for {args.domain}")
                return
            console.print(
                f"Found {len(subdomains)} unique subdomain(s) of {args.domain}"
            )
        print_subdomains(subdomains, args.json)
    except CensysRateLimitExceededException:
        err_console.print("Censys API rate limit exceeded")
        print_subdomains(subdomains, args.json)
    except CensysUnauthorizedException:
        err_console.print("Invalid Censys API ID or secret")
        sys.exit(1)
    except CensysException as e:
        err_console.print(str(e))
        print_subdomains(subdomains, args.json)
        sys.exit(1)
    except KeyboardInterrupt:
        err_console.print("Caught Ctrl-C, exiting...")
        print_subdomains(subdomains, args.json)
        sys.exit(1)


def include(parent_parser: argparse._SubParsersAction, parents: dict):
    """Include this subcommand into the parent parser.

    Args:
        parent_parser (argparse._SubParsersAction): Parent parser.
        parents (dict): Parent arg parsers.
    """
    subdomains_parser = parent_parser.add_parser(
        "subdomains",
        description="Enumerates subdomains using the Censys Search Certificates index",
        help="enumerate subdomains",
        parents=[parents["auth"]],
    )
    subdomains_parser.add_argument("domain", help="The base domain to search for")
    subdomains_parser.add_argument(
        "--pages", type=int, default=1, help="Max records to query"
    )
    subdomains_parser.add_argument(
        "-j", "--json", action="store_true", help="Output in JSON format"
    )
    subdomains_parser.set_defaults(func=cli_subdomains)



"""Interact with the Censys Seeds, Assets, and Logbook APIs."""

from typing import Optional

from .assets import (
    CertificatesAssets,
    DomainsAssets,
    HostsAssets,
    ObjectStoragesAssets,
    SubdomainsAssets,
    WebEntitiesAssets,
)
from .beta import Beta
from .clouds import Clouds
from .inventory import InventorySearch
from .logbook import Logbook
from .risks import Risks
from .saved_queries import SavedQueries
from .seeds import Seeds


class AsmClient:
    """Client ASM API class."""

    def __init__(self, api_key: Optional[str] = None, **kwargs):
        """Inits AsmClient.

        Args:
            api_key (str): Optional; The API Key provided by Censys.
            **kwargs: Arbitrary keyword arguments.
        """
        self.seeds = Seeds(api_key, **kwargs)
        self.hosts = HostsAssets(api_key, **kwargs)
        self.certificates = CertificatesAssets(api_key, **kwargs)
        self.domains = DomainsAssets(api_key, **kwargs)
        self.subdomains = SubdomainsAssets(api_key, **kwargs)
        self.logbook = Logbook(api_key, **kwargs)
        self.events = self.logbook
        self.clouds = Clouds(api_key, **kwargs)
        self.risks = Risks(api_key, **kwargs)
        self.inventory = InventorySearch(api_key, **kwargs)
        self.object_storages = ObjectStoragesAssets(api_key, **kwargs)
        self.web_entities = WebEntitiesAssets(api_key, **kwargs)
        self.beta = Beta(api_key, **kwargs)
        self.saved_queries = SavedQueries(api_key, **kwargs)



"""Base for interacting with the Censys Search API."""

import os
from typing import List, Optional, Type

from requests.models import Response

from censys.common.base import CensysAPIBase
from censys.common.config import DEFAULT, get_config
from censys.common.exceptions import (
    CensysException,
    CensysExceptionMapper,
    CensysSearchException,
)

Fields = Optional[List[str]]


class CensysSearchAPIv1(CensysAPIBase):
    """This class is the base class for all v1 API indexes."""

    DEFAULT_URL: str = "https://search.censys.io/api/v1"
    """Default Search API base URL."""
    INDEX_NAME: Optional[str] = None
    """Name of Censys Index."""

    def __init__(
        self, api_id: Optional[str] = None, api_secret: Optional[str] = None, **kwargs
    ):
        """Inits CensysSearchAPIv1.

        See CensysAPIBase for additional arguments.

        Args:
            api_id (str): Optional; The API ID provided by Censys.
            api_secret (str): Optional; The API secret provided by Censys.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            CensysException: Base Exception Class for the Censys API.
        """
        CensysAPIBase.__init__(self, kwargs.pop("url", self.DEFAULT_URL), **kwargs)

        # Gets config file
        config = get_config()

        # Try to get credentials
        self._api_id = (
            api_id or os.getenv("CENSYS_API_ID") or config.get(DEFAULT, "api_id")
        )
        self._api_secret = (
            api_secret
            or os.getenv("CENSYS_API_SECRET")
            or config.get(DEFAULT, "api_secret")
        )
        if not self._api_id or not self._api_secret:
            raise CensysException("No API ID or API secret configured.")

        self._session.auth = (self._api_id, self._api_secret)

        # Generate concrete paths to be called
        self.search_path = f"/search/{self.INDEX_NAME}"
        self.view_path = f"/view/{self.INDEX_NAME}/"
        self.report_path = f"/report/{self.INDEX_NAME}"

        # Confirm setup
        # self.account()

    def _get_exception_class(  # type: ignore
        self, res: Response
    ) -> Type[CensysSearchException]:
        return CensysExceptionMapper.SEARCH_EXCEPTIONS.get(
            res.status_code, CensysSearchException
        )

    def account(self) -> dict:
        """Gets the current account information.

        This includes email and quota.

        Returns:
            dict: Account response.
        """
        return self._get("account")

    def quota(self) -> dict:
        """Gets the current account's query quota.

        Returns:
            dict: Quota response.
        """
        return self.account()["quota"]
